@@ -15,7 +15,7 @@
 #
 
 package MongoDB::Collection;
-our $VERSION = '0.26';
+our $VERSION = '0.27';
 
 # ABSTRACT: A Mongo Collection
 
@@ -25,7 +25,7 @@ MongoDB::Collection - A Mongo Collection
 
 =head1 VERSION
 
-version 0.26
+version 0.27
 
 =cut
 
@@ -74,6 +74,46 @@ sub _build_full_name {
     return "${db_name}.${name}";
 }
 
+=head1 STATIC METHODS
+
+=head2 to_index_string ($keys)
+
+    $name = MongoDB::Collection::to_index_string({age : 1});
+
+Takes a L<Tie::IxHash>, hash reference, or array reference.  Converts it into
+an index string.
+
+=cut
+
+sub to_index_string {
+    my $keys = shift;
+
+    my @name;
+    if (ref $keys eq 'ARRAY' ||
+        ref $keys eq 'HASH' ) {
+        
+        while ((my $idx, my $d) = each(%$keys)) {
+            push @name, $idx;
+            push @name, $d;
+        }
+    }
+    elsif (ref $keys eq 'Tie::IxHash') {
+        my @ks = $keys->Keys;
+        my @vs = $keys->Values;
+
+        @vs = $keys->Values;
+        for (my $i=0; $i<$keys->Length; $i++) {
+            push @name, $ks[$i];
+            push @name, $vs[$i];
+        }
+    }
+    else {
+        confess 'expected Tie::IxHash, hash, or array reference for keys';
+    }
+
+    return join("_", @name);
+}
+
 =head1 METHODS
 
 =head2 query ($query, \%attrs?)
@@ -118,25 +158,34 @@ Executes the given C<$query> and returns the first object matching it.
 C<$query> can be a hash reference, L<Tie::IxHash>, or array reference (with an
 even number of elements).  
 
-=head2 insert ($object)
+=head2 insert (\%object, \%options?)
 
-    my $id = $collection->insert({ name => 'mongo', type => 'database' });
+    my $id1 = $coll->insert({ name => 'mongo', type => 'database' });
+    my $id2 = $coll->insert({ name => 'mongo', type => 'database' }, {safe => 1});
 
 Inserts the given C<$object> into the database and returns it's id
 value. C<$object> can be a hash reference, a reference to an array with an
 even number of elements, or a L<Tie::IxHash>.  The id is the C<_id> value 
 specified in the data or a L<MongoDB::OID>.
 
-=head2 batch_insert (@array)
+The optional C<$options> parameter can be used to specify if this is a safe 
+insert.  A safe insert will check with the database if the insert succeeded and
+confess that it did not, if an error occured.
+
+=head2 batch_insert (@array, \%options)
 
     my @ids = $collection->batch_insert(({name => "Joe"}, {name => "Fred"}, {name => "Sam"}));
 
 Inserts each of the documents in the array into the database and returns an
 array of their _id fields.
 
+The optional C<$options> parameter can be used to specify if this is a safe 
+insert.  A safe insert will check with the database if the insert succeeded and
+confess that it did not, if an error occured.
+
 =head2 update (\%criteria, \%object, \%options?)
 
-    $collection->update({'x' => 3}, {'$inc' => {'count' => -1}, {"upsert" => 1, "multiple" => 1});
+    $collection->update({'x' => 3}, {'$inc' => {'count' => -1} }, {"upsert" => 1, "multiple" => 1});
 
 Updates an existing C<$object> matching C<$criteria> in the database. 
 
@@ -150,7 +199,8 @@ If no object matching C<$criteria> is found, C<$object> will be inserted.
 
 =item C<multiple>
 All of the documents that match C<$criteria> will be updated, not just 
-the first document found.
+the first document found. (Only available with database version 1.1.3 and 
+newer.)
 
 =back
 
@@ -163,13 +213,33 @@ parameters are given, removes all objects from the collection (but does not
 delete indexes, as C<MongoDB::Collection::drop> does).  Boolean parameter 
 C<$just_one> causes only one matching document to be removed.
 
-=head2 ensure_index ($keys, $direction?, $unique?)
+=head2 ensure_index (\%keys, $options?)
 
-    $collection->ensure_index([qw/foo bar/]);
+    use boolean;
+    $collection->ensure_index({"foo" => 1, "bar" => -1}, { unique => true });
 
-Makes sure the given C<@keys> of this collection are indexed. C<keys> can 
-be an array reference, hash reference, or C<Tie::IxHash>.  The optional
-index direction defaults to C<ascending>.  
+Makes sure the given C<$keys> of this collection are indexed. C<$keys> can be an
+array reference, hash reference, or C<Tie::IxHash>.  C<Tie::IxHash> is prefered
+for multi-key indexes, so that the keys are in the correct order.  1 creates an 
+ascending index, -1 creates a descending index.  
+
+The second parameter gives index options.  Available options are:
+
+=over
+
+=item C<unique => boolean>
+
+By default, indexes are not unique. To create a unique index, pass 
+C<"unique" => true>.  C<true> can be L<boolean::true> or any other true value.
+
+=item C<drop_dups => boolean>
+
+If a unique index is being created on an existing set of data that has duplicate
+values, creating the index will fail.  To force the index creation by deleting 
+duplicate values, use this option.  Again, any value that evaluates to true will
+work.
+
+=back
 
 =cut
 
