@@ -26,13 +26,6 @@
 
 #define PERL_MONGO_CALL_BOOT(name)  perl_mongo_call_xs (aTHX_ name, cv, mark)
 
-/* take care of big endian machines */
-
-#define SERIALIZE(dest, src, len)               \
-  memcpy(dest, src, len);
-
-#define OID_CLASS "MongoDB::OID"
-
 /* whether to add an _id field */
 #define PREP 1
 #define NO_PREP 0
@@ -42,6 +35,30 @@ typedef __int64 int64_t;
 #define inline __inline
 #endif
 
+#if MONGO_BIG_ENDIAN
+#define BYTE1_32(b) ((b & 0xff000000) >> 24)
+#define BYTE2_32(b) ((b & 0x00ff0000) >> 8)
+#define BYTE3_32(b) ((b & 0x0000ff00) << 8)
+#define BYTE4_32(b) ((b & 0x000000ff) << 24)
+#define MONGO_32(b) (BYTE4_32(b) | BYTE3_32(b) | BYTE2_32(b) | BYTE1_32(b))
+
+#define BYTE1_64(b) ((b & 0xff00000000000000ll) >> 56)
+#define BYTE2_64(b) ((b & 0x00ff000000000000ll) >> 40)
+#define BYTE3_64(b) ((b & 0x0000ff0000000000ll) >> 24)
+#define BYTE4_64(b) ((b & 0x000000ff00000000ll) >> 8)
+#define BYTE5_64(b) ((b & 0x00000000ff000000ll) << 8)
+#define BYTE6_64(b) ((b & 0x0000000000ff0000ll) << 24)
+#define BYTE7_64(b) ((b & 0x000000000000ff00ll) << 40)
+#define BYTE8_64(b) ((b & 0x00000000000000ffll) << 56)
+#define MONGO_64(b) (BYTE8_64(b) | BYTE7_64(b) | BYTE6_64(b) | BYTE5_64(b) | BYTE4_64(b) | BYTE3_64(b) | BYTE2_64(b) | BYTE1_64(b))
+
+#else
+#define MONGO_32(b) (b)
+#define MONGO_64(b) (b)
+#endif
+
+
+
 #define INT_32 4
 #define INT_64 8
 #define DOUBLE_64 8
@@ -69,6 +86,7 @@ typedef __int64 int64_t;
 #define BSON_MAXKEY 127
 
 #define GROW_SLOWLY 1048576
+#define MAX_OBJ_SIZE 1024*1024*4
 
 typedef struct {
   char *start;
@@ -97,6 +115,8 @@ void perl_mongo_oid_create(char* twelve, char *twenty4);
 SV *perl_mongo_bson_to_sv (buffer *buf);
 void perl_mongo_sv_to_bson (buffer *buf, SV *sv, AV *ids);
 
+int perl_mongo_resize_buf (buffer*, int);
+void perl_mongo_serialize_key(buffer*, const char*, void*);
 void perl_mongo_serialize_size(char*, buffer*);
 void perl_mongo_serialize_double(buffer*, double);
 void perl_mongo_serialize_string(buffer*, const char*, int);
